use crate::ast::{
    CompilationUnit, DelegateOrIdentifier, Fqn, FunctionModifier, FunctionTypeUse, GenericArgument,
    GenericArguments, GenericParameter, GenericParameters, Identifier, InputArgument,
    InputArguments, InterfaceOrClassTypeUse, Reference, References, ReturnType, TupleTypeUse,
    TypeUse, VoidOrTypeUse,
};
use crate::parser::Rule;
use pest::iterators::Pair;

fn expect_and_use<T>(pair: Pair<Rule>, rule: Rule, f: fn(Pair<Rule>) -> T) -> T {
    if pair.as_rule() != rule {
        panic!("Expected rule {:?} but found {:?}", rule, pair.as_rule())
    }
    f(pair)
}

pub fn build_ast(compilation_unit_pair: Pair<Rule>) -> CompilationUnit {
    todo!()
}

fn build_identifier(identifier_pair: Pair<Rule>) -> Identifier {
    Identifier {
        name: identifier_pair.as_str().to_string(),
    }
}

fn build_fqn(fqn_pair: Pair<Rule>) -> Fqn {
    let mut identifiers: Vec<Identifier> = vec![];
    for identifier_pair in fqn_pair.into_inner() {
        identifiers.push(expect_and_use(
            identifier_pair,
            Rule::Identifier,
            build_identifier,
        ));
    }
    Fqn { identifiers }
}

fn build_type_use(type_use_pair: Pair<Rule>) -> TypeUse {
    let inner_pair = type_use_pair.into_inner().next().unwrap();
    match inner_pair.as_rule() {
        Rule::InterfaceOrClassTypeUse => {
            TypeUse::InterfaceOrClass(build_interface_or_class_type_use(inner_pair))
        }
        Rule::TupleTypeUse => TypeUse::Tuple(build_tuple_type_use(inner_pair)),
        Rule::FunctionTypeUse => TypeUse::Function(build_function_type_use(inner_pair)),
        _ => unreachable!(),
    }
}

fn build_interface_or_class_type_use(pair: Pair<Rule>) -> InterfaceOrClassTypeUse {
    let mut inner = pair.into_inner();
    let fqn = expect_and_use(inner.next().unwrap(), Rule::FullyQualifiedName, build_fqn);

    let generics = inner
        .next()
        .map(|inner_pair| {
            expect_and_use(inner_pair, Rule::GenericArguments, build_generic_arguments)
        })
        .unwrap_or(GenericArguments(vec![]));

    InterfaceOrClassTypeUse { fqn, generics }
}

fn build_tuple_type_use(tuple_type_use_pair: Pair<Rule>) -> TupleTypeUse {
    TupleTypeUse(
        tuple_type_use_pair
            .into_inner()
            .map(|type_use_pair| expect_and_use(type_use_pair, Rule::TypeUse, build_type_use))
            .collect(),
    )
}

fn build_function_type_use(function_pair: Pair<Rule>) -> FunctionTypeUse {
    let mut function_modifier: Option<FunctionModifier> = None;
    let mut generics: Option<GenericParameters> = None;
    let mut parameters: Option<TupleTypeUse> = None;
    let mut inputs: Option<InputArguments> = None;
    let mut return_type: Option<ReturnType> = None;

    for inner_pair in function_pair.into_inner() {
        match inner_pair.as_rule() {
            Rule::FunctionTypeModifier => {
                function_modifier = Some(build_function_modifier(inner_pair));
            }
            Rule::Fn => {}
            Rule::GenericParameters => {
                generics = Some(build_generic_parameters(inner_pair));
            }
            Rule::TupleTypeUse => {
                parameters = Some(build_tuple_type_use(inner_pair));
            }
            Rule::FunctionInputArguments => {
                inputs = Some(build_function_input_arguments(inner_pair));
            }
            Rule::ReturnType => {
                return_type = Some(build_return_type(inner_pair));
            }
            _ => unreachable!(),
        }
    }

    FunctionTypeUse {
        function_modifier,
        generics: generics.unwrap_or(GenericParameters(vec![])),
        parameters: parameters.unwrap(),
        inputs: inputs.unwrap_or(InputArguments(vec![])),
        return_type: return_type.unwrap(),
    }
}

fn build_generic_arguments(generic_arguments_pair: Pair<Rule>) -> GenericArguments {
    let mut generic_arguments: Vec<GenericArgument> = vec![];
    for generic_argument_pair in generic_arguments_pair.into_inner() {
        generic_arguments.push(expect_and_use(
            generic_argument_pair,
            Rule::FullyQualifiedName,
            build_generic_argument,
        ));
    }
    GenericArguments(generic_arguments)
}

fn build_generic_argument(fqn_pair: Pair<Rule>) -> GenericArgument {
    GenericArgument {
        fqn: build_fqn(fqn_pair),
    }
}

fn build_function_modifier(function_modifier_pair: Pair<Rule>) -> FunctionModifier {
    let mut inner = function_modifier_pair.into_inner();
    if inner.len() == 2 {
        FunctionModifier::MutRef
    } else {
        match inner.next().unwrap().as_rule() {
            Rule::Cons => FunctionModifier::Cons,
            Rule::Mut => FunctionModifier::Mut,
            Rule::Ref => FunctionModifier::Ref,
            _ => unreachable!(),
        }
    }
}

fn build_generic_parameters(generic_parameters_pair: Pair<Rule>) -> GenericParameters {
    GenericParameters(
        generic_parameters_pair
            .into_inner()
            .map(|identifier_pair| {
                GenericParameter(expect_and_use(
                    identifier_pair,
                    Rule::Identifier,
                    build_identifier,
                ))
            })
            .collect(),
    )
}

fn build_function_input_arguments(pair: Pair<Rule>) -> InputArguments {
    InputArguments(
        pair.into_inner()
            .map(|function_input_argument_pair| {
                let mut inner = function_input_argument_pair.into_inner();
                let lhs_pair = inner.next().unwrap();
                let lhs = match lhs_pair.as_rule() {
                    Rule::Delegate => DelegateOrIdentifier::Delegate,
                    Rule::Identifier => {
                        DelegateOrIdentifier::Identifier(build_identifier(lhs_pair))
                    }
                    _ => unreachable!(),
                };
                let rhs = build_identifier(inner.next().unwrap());
                InputArgument { lhs, rhs }
            })
            .collect(),
    )
}

fn build_return_type(return_type_pair: Pair<Rule>) -> ReturnType {
    let mut inner = return_type_pair.into_inner();

    let declared_type_pair = inner.next().unwrap();
    let declared_type = match declared_type_pair.as_rule() {
        Rule::Void => VoidOrTypeUse::Void,
        Rule::TypeUse => VoidOrTypeUse::TypeUse(Box::new(build_type_use(declared_type_pair))),
        _ => unreachable!(),
    };

    let references = inner
        .next()
        .map(|ref_list_pair| expect_and_use(ref_list_pair, Rule::RefList, build_references))
        .unwrap_or(References(vec![]));

    ReturnType {
        declared_type,
        references,
    }
}

fn build_references(ref_list_pair: Pair<Rule>) -> References {
    let mut identifiers: Vec<Identifier> = vec![];
    for pair in ref_list_pair.into_inner() {
        match pair.as_rule() {
            Rule::Ref => {}
            Rule::Identifier => {
                identifiers.push(build_identifier(pair));
            }
            _ => unreachable!(),
        }
    }
    References(
        identifiers
            .into_iter()
            .map(|identifier| Reference(identifier))
            .collect(),
    )
}
